// Elevon or V-tail mixer for 16 MHz 328-based types (Uno, Nano, Pro mini etc.)
// two PWM (servo) inputs on the two hardware interrupt pins, 2 and 3, call the pulse widths in microseconds A and B
// four PWM (servo) outputs on pins 4, 5, 6, 7
// these provide A+B, A-B, -A+B, and -A-B mixes respectively
// so there's no need to reverse servos on your transmitter / flight controller
// just plug your two servos into the two outputs that work for your model
// assumes that 'neutral' servo is 1500 us, and limits output ranges 800 - 2200 us
// ceptimus 2022-03-05
// 22-03-06 modified to use accurateMicros() instead of micros()

// #define DEBUG // when not commented out serial debug messages and LED_BUILTIN are active

#ifdef DEBUG
  // time (in microseconds) between debug print events
  #define PRINT_INTERVAL 1000000UL
  uint32_t usDebugPrint = PRINT_INTERVAL;
  int ledToggle = HIGH; // in DEBUG mode, toggle LED on A and B interrupt edges
#endif

// standard servo output update rate is 50 Hz so interval is 20 ms or 20000 us
#define OUTPUT_INTERVAL 20000UL
uint32_t usOutput = OUTPUT_INTERVAL;

int16_t A = 1500, B = 1500;  // the two input channel pulse widths in microseconds (us)

void setup() {
  pinMode(LED_BUILTIN, OUTPUT); // in DEBUG mode, toggle LED on A and B interrupt edges (changes brightness when sticks moved)
  digitalWrite(LED_BUILTIN, LOW); // LED OFF
#ifdef DEBUG
  Serial.begin(115200);
  Serial.print("Debug serial started\n");
#endif
  pinMode(2, INPUT_PULLUP);
  pinMode(3, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(2), A_change, CHANGE);
  attachInterrupt(digitalPinToInterrupt(3), B_change, CHANGE);
  pinMode(4, OUTPUT);
  digitalWrite(4, LOW);
  pinMode(5, OUTPUT);
  digitalWrite(5, LOW);
  pinMode(6, OUTPUT);
  digitalWrite(6, LOW);
  pinMode(7, OUTPUT);
  digitalWrite(7, LOW);
  // using 8-bit Timer2 to provide a more accurate version of micros()  (standard micros() only resloves to 4us)
  TCCR2A = 0x00;
  TCCR2B = 0x02; // /8 clock prescale
  TIMSK2 |= 0x01;
  // using 16-bit Timer1 to time simple intervals in microseconds for outputting the servo pulses
  TCCR1A = 0x00;
  TIMSK1 = 0x01; // enable timer interrupt
}

void loop() {
  if (usOutput - accurateMicros() > 0x80000000UL) { // it's time to output to the servos
    TCNT1 = ~(constrain(A + B - 1500, 800, 2200) * 2); // pulse width for first servo output in half-microseconds
    PORTD |= 0b10000; // pin 4 high
    TCCR1B = 0x02; // start Timer1 running in half-microsecond units
    while (PIND & 0b10000) continue; // wait for timer to pull pin low
    TCNT1 = ~(constrain(A - B + 1500, 800, 2200) * 2); // similarly for 2nd, 3rd, and 4th servo outputs
    PORTD |= 0b100000; // pin 5 high
    TCCR1B = 0x02;
    while (PIND & 0b100000) continue;
    TCNT1 = ~(constrain(-A + B + 1500, 800, 2200) * 2);
    PORTD |= 0b1000000; // pin 6 high
    TCCR1B = 0x02;
    while (PIND & 0b1000000) continue;
    TCNT1 = ~(constrain(-A - B + 4500, 800, 2200) * 2);
    PORTD |= 0b10000000; // pin 7 high
    TCCR1B = 0x02;
    while (PIND & 0b10000000) continue;
    usOutput += OUTPUT_INTERVAL;
  }
#ifdef DEBUG
  if (usDebugPrint - accurateMicros() > 0x80000000UL) { // it's time to print
    Serial.print("\nA = ");
    Serial.print(A);
    Serial.print("\tB = ");
    Serial.print(B);
    Serial.print("\tA+B = ");
    Serial.print(constrain(A + B - 1500, 800, 2200));
    Serial.print("\tA-B = ");
    Serial.print(constrain(A - B + 1500, 800, 2200));
    Serial.print("\t-A+B = ");
    Serial.print(constrain(-A + B + 1500, 800, 2200));
    Serial.print("\t-A-B = ");
    Serial.print(constrain(-A - B + 4500, 800, 2200));
    usDebugPrint += PRINT_INTERVAL;
  }
#endif
}

void A_change(void) { // Input A (pin 2) interrupt
  static uint32_t previousMicros = 0UL;
  uint32_t currentMicros = accurateMicros();
  if (PIND & 0b100) { // rising edge - just remember the timestamp
    previousMicros = currentMicros;
  } else { // falling edge: PWM for A is current time - previous time
    A = (int16_t)(currentMicros - previousMicros);
  }
#ifdef DEBUG
  digitalWrite(LED_BUILTIN, ledToggle);
  if (ledToggle == HIGH) {
    ledToggle = LOW;
  } else {
    ledToggle = HIGH;
  }
#endif
}

void B_change(void) { // Input B (pin 3) interrupt
  static uint32_t previousMicros = 0UL;
  uint32_t currentMicros = accurateMicros();
  if (PIND & 0b1000) { // rising edge - just remember the timestamp
    previousMicros = currentMicros;
  } else { // falling edge: PWM for B is current time - previous time
    B = (int16_t)(currentMicros - previousMicros);
  }
#ifdef DEBUG
  digitalWrite(LED_BUILTIN, ledToggle);
  if (ledToggle == HIGH) {
    ledToggle = LOW;
  } else {
    ledToggle = HIGH;
  }
#endif
}

ISR(TIMER1_OVF_vect) { // Timer1 interrupt
  PORTD &= 0b00001111; // terminate any of the four servo pulses by pulling all four output pins low.  Only one will be high
  TCCR1B = 0x00; // stop Timer1
}

// variables and code to provide accurateMicros() replacement for micros() with four times better resolution
volatile uint32_t timer2cycles = 0UL;
ISR(TIMER2_OVF_vect) {
  ++timer2cycles;
}

uint32_t accurateMicros(void) {
  uint8_t sreg = SREG;
  cli();
  uint32_t u = timer2cycles;
  uint8_t t = TCNT2;
  if (TIFR2 & 0x01) {
    t = TCNT2;
    timer2cycles = ++u;
    TIFR2 |= 0x01;
  }
  SREG = sreg;
  return F_CPU <= 8000000 ? (u << 8) | t : (u << 7) | (t >> 1);
}
